#include "ap.h"


inline size_t min (size_t a, size_t b)
{
    return (a < b ? a : b);
}

// Highly parallel three modular multiplications using the FPU
inline void modmul3 (modint &a1, modint &b1, modint &a2, modint &b2, modint &a3, modint &b3)
{
    asm ("fildl %3; fildl %4;
          fildl %5; fxch %%st(2);
          fmulp %%st, %%st(1); fxch %%st(1);
          fildl %6; fildl %7; fxch %%st(2);
          fmulp %%st, %%st(1); fxch %%st(1);
          fildl %8; fmulp %%st, %%st(1);
          fld %%st(2); fmul %%st(4), %%st;
          fld %%st(2); fmul %%st(5), %%st;
          fld %%st(2); fmul %%st(6), %%st; fxch %%st(2);
          fadds _chopper64; fxch %%st(1);
          fadds _chopper64; fxch %%st(2);
          fadds _chopper64; fxch %%st(1);
          fsubs _chopper64; fxch %%st(2);
          fsubs _chopper64; fxch %%st(1);
          fsubs _chopper64; fxch %%st(2);
          fmull _dmodulus; fxch %%st(1);
          fmull _dmodulus; fxch %%st(1);
          fsubrp %%st, %%st(5); fxch %%st(1);
          fmull _dmodulus; fxch %%st(1);
          fsubrp %%st, %%st(3);
          fsubrp %%st, %%st(1); fxch %%st(2);
          fistpl %0;
          fistpl %1;
          fistpl %2"
                            : "=m" (a1), "=m" (a2), "=m" (a3)
                            : "0" (a1), "m" (b1), "1" (a2), "m" (b2), "2" (a3), "m" (b3)
                            : "cc");
}

// Partially parallel two modular multiplications using the FPU
inline void modmul2 (modint &a1, modint &b1, modint &a2, modint &b2)
{
    asm ("fildl %2; fildl %3;
          fildl %4; fxch %%st(2);
          fmulp %%st, %%st(1); fxch %%st(1);
          fildl %5; fmulp %%st, %%st(1);
          fld %%st(1); fmul %%st(3), %%st;
          fld %%st(1); fmul %%st(4), %%st; fxch %%st(1);
          fadds _chopper64; fxch %%st(1);
          fadds _chopper64; fxch %%st(1);
          fsubs _chopper64; fxch %%st(1);
          fsubs _chopper64; fxch %%st(1);
          fmull _dmodulus; fxch %%st(1);
          fmull _dmodulus; fxch %%st(1);
          fsubrp %%st, %%st(3);
          fsubrp %%st, %%st(1); fxch %%st(1);
          fistpl %0;
          fistpl %1"
                            : "=m" (a1), "=m" (a2)
                            : "0" (a1), "m" (b1), "1" (a2), "m" (b2)
                            : "cc");
}

// The "six-step" fnt, but doesn't transpose or scramble (for convolution only)

void tablesixstepfnttrans2 (modint data[], modint pr, int isign, size_t nn)
{
    size_t n1, n2, j, k;
    modint w, tmp, tmp2, *p1, *p2;

    if (nn < 2) return;

    for (n1 = 1, n2 = 0; n1 < nn; n1 <<= 1, n2++);
    n1 = n2 >> 1;
    n2 -= n1;

    n1 = 1 << n1;
    n2 = 1 << n2;

    // n2 >= n1

    modint *wtable = new modint[n2];
    size_t *ptable = new size_t[n1];

    if (isign > 0)
        w = pow (pr, modint::modulus - 1 - (modint::modulus - 1) / nn);
    else
        w = pow (pr, (modint::modulus - 1) / nn);

    // treat the input data as a n1 x n2 matrix

    // first transpose the matrix

    transpose (data, n1, n2);

    // then do n2 transforms of length n1

    // init tables

    tmp = pow (w, nn / n1);
    tmp2 = 1;
    for (k = 0; k < n1; k++)
    {
        wtable[k] = tmp2;
        tmp2 *= tmp;
    }

    initscrambletable (ptable, n1);

    for (k = 0, p1 = data; k < n2; k++, p1 += n1)
        tablefnt (p1, wtable, ptable, n1);

    // transpose the matrix

    transpose (data, n2, n1);

    // then multiply the matrix A_jk by exp(isign * -2 pi i j k / nn)

    tmp = w;
    for (j = 1, p1 = data + n2; j < n1; j++, p1 += n2)
    {
        tmp2 = pow (tmp, j);
        p1[j] *= tmp2;
        tmp2 *= tmp;
        for (k = j + 1, p2 = p1 + n2 + j; k < n1; k++, p2 += n2)
        {
            /*
            p1[k] *= tmp2;
            *p2 *= tmp2;
            tmp2 *= tmp;
            */

            modmul3 (tmp2, tmp, p1[k], tmp2, *p2, tmp2);
        }
        for (; k < n2; k++)
        {
            /*
            p1[k] *= tmp2;
            tmp2 *= tmp;
            */

            modmul2 (tmp2, tmp, p1[k], tmp2);
        }
        tmp *= w;
    }

    // last do n1 transforms of length n2

    // init table

    if (n2 != n1)
    {
        // n2 = 2 * n1
        for (k = n1; k--;)
            wtable[2 * k] = wtable[k];
        tmp = pow (w, nn / n2);
        tmp2 = tmp;
        tmp *= tmp;
        for (k = 1; k < n2; k += 2)
        {
            wtable[k] = tmp2;
            tmp2 *= tmp;
        }
    }

    for (k = 0, p1 = data; k < n1; k++, p1 += n2)
        tablefnt (p1, wtable, 0, n2, 0);

    delete[] ptable;
    delete[] wtable;
}

void tablesixstepfnttrans (modint data[], modint pr, int isign, size_t nn)
{
    size_t n2 = (nn & -nn), j, k, s;
    modint w, ww, w1, w2, w3, *p1, *p2, *p3, tmp, tmp2, tmp3, *d, t;

    if (nn < 2) return;

    if (nn == n2)
    {
        // Transform length is a power of two
        tablesixstepfnttrans2 (data, pr, isign, nn);
        return;
    }

    // Transform length is three times a power of two

    if (isign > 0)
        w = pow (pr, modint::modulus - 1 - (modint::modulus - 1) / nn);
    else
        w = pow (pr, (modint::modulus - 1) / nn);

    ww = w * w;

    w3 = pow (w, n2);                   // 3rd root of unity
    w1 = -modint (3) / modint (2);
    w2 = w3 + modint (1) / modint (2);

    s = min (n2, Cachemaxblocksize / 4);
    d = new modint[3 * s];

    tmp = tmp2 = 1;
    for (k = 0; k < n2; k += s)
    {
        p1 = d;
        p2 = p1 + s;
        p3 = p2 + s;
        moveraw (p1, data + k, s);                      // Cache in
        moveraw (p2, data + k + n2, s);
        moveraw (p3, data + k + 2 * n2, s);
        for (j = 0; j < s; j++, p1++, p2++, p3++)
        {
            t = *p2 + *p3;                              // Transform columns
            *p3 = *p2 - *p3;
            *p1 += t;
            tmp3 = tmp;
            modmul3 (t, w1, *p3, w2, tmp, w);
            t += *p1;
            *p2 = t + *p3;
            *p3 = t - *p3;
            modmul3 (*p2, tmp3, *p3, tmp2, tmp2, ww);   // Multiply
        }
        p1 = d;
        p2 = p1 + s;
        p3 = p2 + s;
        moveraw (data + k, p1, s);                      // Cache out
        moveraw (data + k + n2, p2, s);
        moveraw (data + k + 2 * n2, p3, s);
    }

    delete[] d;

    tablesixstepfnttrans2 (data, pr, isign, n2);        // Transform rows
    tablesixstepfnttrans2 (data + n2, pr, isign, n2);
    tablesixstepfnttrans2 (data + 2 * n2, pr, isign, n2);
}

void itablesixstepfnttrans2 (modint data[], modint pr, int isign, size_t nn, size_t e)
{
    size_t n1, n2, j, k;
    modint w, tmp, tmp2, *p1, *p2, inn;

    if (nn < 2) return;

    for (n1 = 1, n2 = 0; n1 < nn; n1 <<= 1, n2++);
    n1 = n2 >> 1;
    n2 -= n1;

    n1 = 1 << n1;
    n2 = 1 << n2;

    // n2 >= n1

    modint *wtable = new modint[n2];
    size_t *ptable = new size_t[n1];

    if (isign > 0)
        w = pow (pr, modint::modulus - 1 - (modint::modulus - 1) / nn);
    else
        w = pow (pr, (modint::modulus - 1) / nn);

    // treat the input data as a n1 x n2 matrix

    // first do n1 transforms of length n2

    // init table

    tmp = pow (w, nn / n2);
    tmp2 = 1;
    for (k = 0; k < n2; k++)
    {
        wtable[k] = tmp2;
        tmp2 *= tmp;
    }

    for (k = 0, p1 = data; k < n1; k++, p1 += n2)
        itablefnt (p1, wtable, 0, n2, 0);

    // then multiply the matrix A_jk by exp(isign * -2 pi i j k / nn) / nn

    tmp = 1;
    inn = modint (1) / modint (nn * e);
    for (j = 0, p1 = data; j < n1; j++, p1 += n2)
    {
        tmp2 = pow (tmp, j) * inn;
        p1[j] *= tmp2;
        tmp2 *= tmp;
        for (k = j + 1, p2 = p1 + n2 + j; k < n1; k++, p2 += n2)
        {
            /*
            p1[k] *= tmp2;
            *p2 *= tmp2;
            tmp2 *= tmp;
            */

            modmul3 (tmp2, tmp, p1[k], tmp2, *p2, tmp2);
        }
        for (; k < n2; k++)
        {
            /*
            p1[k] *= tmp2;
            tmp2 *= tmp;
            */

            modmul2 (tmp2, tmp, p1[k], tmp2);
        }
        tmp *= w;
    }

    // transpose the matrix

    transpose (data, n1, n2);

    // then do n2 transforms of length n1

    // init table

    if (n2 != n1)
        // n2 = 2 * n1
        for (k = 0; k < n1; k++)
            wtable[k] = wtable[2 * k];

    initscrambletable (ptable, n1);

    for (k = 0, p1 = data; k < n2; k++, p1 += n1)
        itablefnt (p1, wtable, ptable, n1);

    // last transpose the matrix

    transpose (data, n2, n1);

    delete[] ptable;
    delete[] wtable;
}

void itablesixstepfnttrans (modint data[], modint pr, int isign, size_t nn)
{
    size_t n2 = (nn & -nn), j, k, s;
    modint w, ww, w1, w2, w3, *p1, *p2, *p3, tmp, tmp2, *d, t;

    if (nn < 2) return;

    if (nn == n2)
    {
        // Transform length is a power of two
        itablesixstepfnttrans2 (data, pr, isign, nn);
        return;
    }

    // Transform length is three times a power of two

    if (isign > 0)
        w = pow (pr, modint::modulus - 1 - (modint::modulus - 1) / nn);
    else
        w = pow (pr, (modint::modulus - 1) / nn);

    itablesixstepfnttrans2 (data, pr, isign, n2, 3);    // Transform rows
    itablesixstepfnttrans2 (data + n2, pr, isign, n2, 3);
    itablesixstepfnttrans2 (data + 2 * n2, pr, isign, n2, 3);

    ww = w * w;

    w3 = pow (w, n2);                   // 3rd root of unity
    w1 = -modint (3) / modint (2);
    w2 = w3 + modint (1) / modint (2);

    s = min (n2, Cachemaxblocksize / 4);
    d = new modint[3 * s];

    tmp = tmp2 = 1;
    for (k = 0; k < n2; k += s)
    {
        p1 = d;
        p2 = p1 + s;
        p3 = p2 + s;
        moveraw (p1, data + k, s);                      // Cache in
        moveraw (p2, data + k + n2, s);
        moveraw (p3, data + k + 2 * n2, s);
        for (j = 0; j < s; j++, p1++, p2++, p3++)
        {
            modmul3 (*p2, tmp, *p3, tmp2, tmp, w);      // Multiply
            t = *p2 + *p3;                              // Transform columns
            *p3 = *p2 - *p3;
            *p1 += t;
            modmul3 (t, w1, *p3, w2, tmp2, ww);
            t += *p1;
            *p2 = t + *p3;
            *p3 = t - *p3;
        }
        p1 = d;
        p2 = p1 + s;
        p3 = p2 + s;
        moveraw (data + k, p1, s);                      // Cache out
        moveraw (data + k + n2, p2, s);
        moveraw (data + k + 2 * n2, p3, s);
    }

    delete[] d;
}
